/*
Copyright (C) 2003-2004 Douglas Thain and the University of Wisconsin
Copyright (C) 2005- The University of Notre Dame
This software is distributed under a BSD-style license.
See the file COPYING for details.
*/

/*
chirp_distribute copies a directory from one host to many others
by building a spanning tree at runtime using third party transfers.
The -X option will delete the directory from all of the named hosts.
*/

#include "chirp_reli.h"

#include "auth.h"
#include "auth_all.h"
#include "debug.h"
#include "stringtools.h"
#include "timestamp.h"
#include "macros.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>
#include <string.h>

static int timeout = 300;
static int overall_timeout = 3600;
static int overall_stoptime = 0;
static int did_explicit_auth = 0;
static int destroy_mode = 0;
static int detail_mode = 0;
static int maxprocs = 100;
static int randomize_mode = 0;

static char *failure_matrix = 0;
static int failure_matrix_size = 0;
static char *failure_matrix_filename = 0;

static int compute_stoptime() 
{
	return MIN(time(0)+timeout,overall_stoptime);
}

#define FAILURE_MARK_NONE    ' '
#define FAILURE_MARK_FAILED  '#'
#define FAILURE_MARK_SUCCESS '+'

static void failure_matrix_init( int n )
{
	failure_matrix = malloc(sizeof(char)*n*n);
	memset(failure_matrix,FAILURE_MARK_NONE,sizeof(char)*n*n);
	failure_matrix_size = n;
}

static void failure_matrix_set( int s, int t, char c )
{
	failure_matrix[s*failure_matrix_size+t] = c;
}

static char failure_matrix_get( int s, int t )
{
	return failure_matrix[s*failure_matrix_size+t];
}

static void failure_matrix_print()
{
	FILE *file;
	int i,j;

	if(!failure_matrix_filename) return;

	file = fopen(failure_matrix_filename,"w");
	if(!file) return;

	for(i=0;i<failure_matrix_size;i++) {
		for(j=0;j<failure_matrix_size;j++) {
			fputc(failure_matrix_get(i,j),file);
		}
		fputc('\n',file);
	}

	fclose(file);
}

typedef enum {
	TARGET_STATE_FRESH,
	TARGET_STATE_RECEIVING,
	TARGET_STATE_SENDING,
	TARGET_STATE_IDLE,
	TARGET_STATE_FAILED
} target_state_t;

struct target_info {
	const char *name;
	target_state_t state;
	pid_t pid;
};

static void show_version( const char *cmd )
{
	printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_use()
{
	printf("Use: chirp_distribute [options] <sourcehost> <sourcepath> <host1> <host2> ...\n");
	printf("where options are:\n");
	printf(" -X         Delete data from all of the target hosts.\n");
	printf(" -D         Show detailed locatin, time, and performance of each transfer.\n");
	printf(" -F <file>  Write matrix of failures to this file.\n");
	printf(" -T <time>  Overall timeout for entire distribution. (default is %d)\n",overall_timeout);
	printf(" -R         Randomize order of target hosts given on command line.\n");
	printf(" -t <time>  Timeout for for each copy. (default is %ds)\n",timeout);
	printf(" -p <num>   Maximum number of processes to run at once (default=%d)\n",maxprocs);
	printf(" -a <flag>  Require this authentication mode.\n");
	printf(" -d <flag>  Enable debugging for this subsystem.\n");
	printf(" -v         Show program version.\n");
	printf(" -h         This message.\n");
	printf("\nchirp_distribute copies a directory from one host to many hosts\n");
	printf("by creating a spanning tree and then transferring data in parallel\n");
	printf("using third party transfer.  The path of each newly created copy\n");
	printf("is printed on stdout.  The -X option deletes all but one copy.\n\n");
}

int main( int argc, char *argv[] )
{
	int result;
	char c;
	char *sourcehost, *sourcepath;
	struct target_info *targets;
	int ntargets;
	int i;
	pid_t pid;
	int nprocs=0;

	srand(time(0));

	debug_config(argv[0]);

	while(((c=getopt(argc,argv,"RXDF:T:p:d:vt:a:h"))!=(char)-1)) {
		switch(c) {
			case 'R':
				randomize_mode = 1;
				break;
			case 'X':
				destroy_mode = 1;
				break;
			case 'D':
				detail_mode = 1;
				break;
			case 'F':
				failure_matrix_filename = optarg;
				break;
			case 'T':
				overall_timeout = string_time_parse(optarg);
				break;
			case 'p':
				maxprocs = atoi(optarg);
				break;
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'v':
				show_version(argv[0]);
				return 0;
				break;
			case 't':
				timeout = string_time_parse(optarg);
				break;
			case 'a':
				auth_register_byname(optarg);
				did_explicit_auth = 1;
				break;
			default:
			case 'h':
				show_use();
				return 1;
		}
	}

	if(!did_explicit_auth) auth_register_all();

	if((argc-optind)<2) {
		show_use();
		return 1;
	}

	sourcehost = argv[optind];
	sourcepath = argv[optind+1];
	ntargets = argc-optind-1;

	if(randomize_mode) {
		for(i=0;i<ntargets*10;i++) {
			int a = optind+2+rand()%(ntargets-1);
			int b = optind+2+rand()%(ntargets-1);
			char *t;

			t = argv[a];
			argv[a] = argv[b];
			argv[b] = t;
		}
	}

	overall_stoptime = time(0) + overall_timeout;

	failure_matrix_init(ntargets);

	if(destroy_mode) {
		int result;
		int status;

		for(i=0;i<(ntargets-1);i++) {
			if(nprocs<maxprocs) {
				nprocs++;
				fflush(0);
				pid = fork();
				if(pid>0) {
					/* keep going */
				} else if(pid==0) {
					result = chirp_reli_rmall(argv[optind+2+i],sourcepath,compute_stoptime());
					if(result==0) {
						printf("destroyed %s\n",argv[optind+2+i]);
						exit(0);
					} else {
						printf("couldn't destroy %s: %s\n",argv[optind+2+i],strerror(errno));
						exit(1);
					}
				} else {
					printf("couldn't fork: %s\n",strerror(errno));
					sleep(1);
					i--;
				}
			} else {
				pid = wait(&status);
				if(pid>0) nprocs--;
				i--;
			}
		}

		while(1) {
			pid = wait(&status);
			if(pid<0 && errno==ECHILD) break;
		}

		return 0;
	}

	targets = malloc(sizeof(struct target_info)*ntargets);

	targets[0].name = sourcehost;
	targets[0].state = TARGET_STATE_IDLE;

	for(i=1;i<ntargets;i++) {
		targets[i].name = argv[optind+1+i];
		targets[i].state = TARGET_STATE_FRESH;
	}

	if(detail_mode) {
		printf("%u   start -> %s    0 secs, 0 MB/sec\n",(unsigned)time(0),sourcehost);
	}

	while(time(0)<overall_stoptime) {
		int source = -1;
		int target = -1;

		for(i=(ntargets-1);i>=0;i--) {
			if(targets[i].state==TARGET_STATE_IDLE) {
				source = i;
				break;
			}
		}

		if(source!=-1) {
			int freshtargets = 0;
			for(i=0;i<ntargets;i++) {
				if(targets[i].state==TARGET_STATE_FRESH) {
					freshtargets++;
					if(failure_matrix_get(source,i)!=FAILURE_MARK_FAILED) {
						target = i;
						break;
					}
				}
			}

			if(target==-1 && freshtargets>0) {
				/* nothing left for this one to do */
				targets[source].state = TARGET_STATE_FAILED;
				continue;
			}
		}

		if(nprocs<maxprocs && target!=-1 && source!=-1) {
			fflush(0);
			pid = fork();
			if(pid>0) {
				nprocs++;
				targets[source].state = TARGET_STATE_SENDING;
				targets[source].pid = pid;
				targets[target].state = TARGET_STATE_RECEIVING;
				targets[target].pid = pid;
			} else if(pid==0) {
				timestamp_t start,stop;
				start = timestamp_get();
				result = chirp_reli_thirdput(targets[source].name,sourcepath,targets[target].name,sourcepath,compute_stoptime());
				stop = timestamp_get();
				if(start==stop) stop++;
				if(result>=0) {
					if(detail_mode) {
						printf("%u   %s -> %s    %.2lf secs, %.1lf MB/sec\n",(unsigned)time(0),targets[source].name,targets[target].name,(stop-start)/1000000.0,result/(double)(stop-start));
					} else {
						printf("/chirp/%s/%s\n",targets[target].name,sourcepath);
					}
					fflush(0);
					exit(0);
				} else {
					int save_errno = errno;
					printf("%u   %s -> %s    failed: %s\n",(unsigned)time(0),targets[source].name,targets[target].name,strerror(errno));
					fflush(0);
					chirp_reli_rmall(targets[target].name,sourcepath,compute_stoptime());
					exit(save_errno);
				}
			} else {
				fprintf(stderr,"chirp_distribute: %s\n",strerror(errno));
				sleep(1);
			}
		} else {
			int status;
			pid = wait(&status);
			if(pid>0) {
				int transfer_ok;
				int error_code;

				nprocs--;

				if(WIFEXITED(status)) {
					error_code = WEXITSTATUS(status);
					transfer_ok = (error_code==0);
				} else {
					transfer_ok = 0;
					error_code = 0;
				}

				for(i=0;i<ntargets;i++) {
					if( targets[i].pid==pid ) {
						if(targets[i].state==TARGET_STATE_RECEIVING) {
							target = i;
							if(transfer_ok) {
								targets[i].state = TARGET_STATE_IDLE;
							} else if(error_code==ECONNRESET) {
								targets[i].state = TARGET_STATE_FAILED;
							} else {
								targets[i].state = TARGET_STATE_FRESH;
							}
						} else if(targets[i].state==TARGET_STATE_SENDING) {
							source = i;
							targets[i].state = TARGET_STATE_IDLE;
						}
					}
				}
				if(transfer_ok) {
					failure_matrix_set(source,target,FAILURE_MARK_SUCCESS);
				} else {
					failure_matrix_set(source,target,FAILURE_MARK_FAILED);
				}
			} else if(errno==ECHILD) {
				break;
			} else {
				fprintf(stderr,"chirp_distribute: wait: %s\n",strerror(errno));
				sleep(1);
			}
		}
	}

	failure_matrix_print();

	return 0;
}
